from typing import Any, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.axes import Axes
from pyspark.sql import DataFrame


def _format_spark_dataframe_for_visualization(
    spark_dataframe: DataFrame,
) -> List[Dict[str, Any]]:
    """
    Prepare a Pandas DataFrame to be displayed as a heatmap visualizing the group and batch counts.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame with `group` and `batch` columns.

    Returns
    -------
    List[Dict[str, Any]]
        The Spark DataFrame contents processed and formatted as a list of dictionaries.
    """
    counts_sdf = spark_dataframe.groupBy("group", "batch").count()
    return [row.asDict() for row in counts_sdf.collect()]


def create_ingest_heatmap(
    spark_dataframe: DataFrame,
    title: str = "Parallel Ingest Heat Map",
    figsize: Optional[Tuple[float, float]] = None,
) -> Axes:
    """
    Create the ingest heatmap from a list of dictionaries.
    This heatmap will display batches on the y-axis and group numbers on the x-axis.
    Group IDs will be displayed in parenthesis below the value count in each cell.

    Parameters
    ----------
    spark_dataframe : DataFrame
        A Spark DataFrame with columns including 'group', 'batch' and 'count'
    title : str, optional
        A title for the visualization, by default "Parallel Ingest Heat Map"
    figsize : tuple, optional
        Figure size (width, height) in inches, by default None

    Returns
    -------
    Axes
        A Matplotlib Axes object for visualization.
    """
    data = _format_spark_dataframe_for_visualization(spark_dataframe=spark_dataframe)

    assert (
        set(data[0].keys()) == {"group", "batch", "count"}
    ), "Invalid keys detected in data. Dictionary keys must contain only 'group', 'batch' and 'count'"

    # Create a dictionary to store group-to-number mapping for each batch
    batch_group_mappings = {}
    for d in data:
        batch = d["batch"]
        if batch not in batch_group_mappings:
            batch_group_mappings[batch] = {}
        batch_group_mappings[batch][d["group"]] = len(batch_group_mappings[batch]) + 1

    # Transform data with group numbers
    transformed_data = []
    for d in data:
        transformed_data.append(
            {
                "batch": d["batch"],
                "group_num": batch_group_mappings[d["batch"]][d["group"]],
                "count": d["count"],
                "original_group": d["group"],
            }
        )

    # Extract unique x and y values
    y_values = sorted(set(d["batch"] for d in transformed_data), reverse=True)
    x_values = sorted(set(d["group_num"] for d in transformed_data))

    # Create 2D numpy arrays for the heatmap
    heatmap_data = np.zeros((len(y_values), len(x_values)))
    annotation_labels = np.empty((len(y_values), len(x_values)), dtype=object)

    # Fill the arrays with values
    for item in transformed_data:
        y_idx = y_values.index(item["batch"])
        x_idx = x_values.index(item["group_num"])
        heatmap_data[y_idx, x_idx] = item["count"]
        # Create annotation with count and original group name
        annotation_labels[y_idx, x_idx] = (
            f"{item['count']:,.0f}\n({item['original_group']})"
        )

    # Create figure with specified size
    plt.figure(figsize=figsize)

    # Create heatmap
    ax = sns.heatmap(
        data=heatmap_data,
        annot=annotation_labels,
        fmt="",
        xticklabels=x_values,
        yticklabels=y_values,
        linewidths=0.5,
        vmin=0,
    )

    ax.set_xlabel("Spark Processor Node Number")
    ax.set_ylabel("Batch")
    ax.set_title(title)

    return ax



from typing import List, Optional

from pyspark.sql import DataFrame

from neo4j_parallel_spark_loader import ingest_spark_dataframe
from neo4j_parallel_spark_loader.bipartite import (
    group_and_batch_spark_dataframe as group_and_batch_bipartite,
)
from neo4j_parallel_spark_loader.predefined_components import (
    group_and_batch_spark_dataframe as group_and_batch_predefined,
)


def build_relationship(
    df: DataFrame,
    relationship_name: str,
    source_labels: str,
    source_keys: str,
    target_labels: str,
    target_keys: str,
    rel_props: List[str] = None,
    group_keys: List[str] = None,
    num_groups: Optional[int] = 10,
    max_serial: Optional[int] = 1000000,
) -> None:
    """Build a relationship between two nodes.
    Params:
        rel_props: List[str]
            list of df columns to use as rel properties
        group_keys: List[str]
            list of df columns to use as group keys for parallel processing.
            1 key uses predefined batching, 2 uses bipartite batching
        num_groups: Optional[int] , optional
            The number of partitions to split Spark DataFrame into. By default 10
        max_serial: Optional[int] , optional
            The maximum number of relationships to process serially.
            Any number of rows above this number will be processed in parallel
    """
    options = {
        "relationship": relationship_name,
        "relationship.save.strategy": "keys",
        "relationship.source.save.mode": "Match",
        "relationship.source.labels": source_labels,
        "relationship.source.node.keys": source_keys,
        "relationship.target.save.mode": "Match",
        "relationship.target.labels": target_labels,
        "relationship.target.node.keys": target_keys,
    }
    if rel_props:
        options["relationship.properties"] = ",".join(rel_props)

    print(f"""Building {df.count()} relationships""")
    if group_keys and len(group_keys) > 0 and df.count() > max_serial:
        print("Building in parallel")
        if len(group_keys) == 1:
            print("Using Predefined Grouping")
            batched_df = group_and_batch_predefined(df, group_keys[0], num_groups)
        else:
            print("Using Bipartite Grouping")
            batched_df = group_and_batch_bipartite(
                df, group_keys[0], group_keys[1], num_groups
            )

        ingest_spark_dataframe(
            spark_dataframe=batched_df,
            save_mode="Overwrite",
            options=options,
            num_groups=num_groups,
        )
    else:
        print("Building in series")
        df = (
            df.write.format("org.neo4j.spark.DataSource")
            .mode("Overwrite")
            .options(**options)
            .save()
        )



from .batching import create_ingest_batches_from_groups
from .grouping import create_node_groupings
from .grouping_and_batching import group_and_batch_spark_dataframe

__all__ = [
    "create_ingest_batches_from_groups",
    "create_node_groupings",
    "group_and_batch_spark_dataframe",
]



from typing import Dict

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col, concat, lit


def create_value_groupings(
    value_counts_spark_dataframe: DataFrame,
    num_groups: int,
    grouping_column: str,
):
    """
    Create a Spark DataFrame containing groupings according to property value counts.

    Parameters
    ----------
    value_counts_spark_dataframe : DataFrame
        The Spark DataFrame containing value counts under the `count` column.
    num_groups : int
        The number of groups to create.
    grouping_column : str
        The column that `count` column refers to.

    Returns
    -------
    DataFrame
        A Spark DataFrame containing columns `value` and `group`.
    """
    spark: SparkSession = value_counts_spark_dataframe.sparkSession
    # to create buckets
    # # track with 2 separate hash maps
    counts_bucket = {i: 0 for i in range(num_groups)}
    keys_bucket = {i: list() for i in range(num_groups)}
    # stack source and target
    # group by and count

    counts_df = value_counts_spark_dataframe.orderBy("count", ascending=False)
    # iterate through the values in max -> min order ex: [{key: Amazon, value_count: 100000}, ...]
    # find most-empty bucket (num_groups) and place value in it and increment bucket value by value_count
    for row in counts_df.collect():
        smallest_bucket_id = _get_smallest_bucket_id(counts_bucket)
        counts_bucket[smallest_bucket_id] = (
            counts_bucket.get(smallest_bucket_id) + row["count"]
        )
        keys_bucket.get(smallest_bucket_id).append(row[grouping_column])

    key_to_group_map = list()
    for bucket_id, lst in keys_bucket.items():
        for v in lst:
            key_to_group_map.append({"value": v, "group": bucket_id})

    return spark.createDataFrame(key_to_group_map)


def _get_smallest_bucket_id(bucket: Dict[int, int]) -> int:
    """Return the key of the smallest bucket."""

    min_val = float("inf")
    min_bucket_id = 0
    for k, v in bucket.items():
        if v < min_val:
            min_val = v
            min_bucket_id = k

    return min_bucket_id


def create_group_column_from_source_and_target_groups(
    spark_dataframe: DataFrame,
) -> DataFrame:
    """Add a `group` column to the Spark DataFrame."""

    return spark_dataframe.withColumn(
        "group", concat(col("source_group"), lit(" --> "), col("target_group"))
    )


def create_value_counts_dataframe(
    spark_dataframe: DataFrame, grouping_column: str
) -> DataFrame:
    """
    Create a `count` column based on the `grouping_column` argument.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame to operate on.
    grouping_column : str
        The grouping column to use.

    Returns
    -------
    DataFrame
        The value counts Spark DataFrame.
    """
    sdf_filtered = spark_dataframe.select(grouping_column)

    counts_df: DataFrame = sdf_filtered.groupBy(grouping_column).count()

    return counts_df



from pyspark.sql import DataFrame

from ..utils.grouping import (
    create_group_column_from_source_and_target_groups,
    create_value_counts_dataframe,
    create_value_groupings,
)
from ..utils.verify_spark import verify_spark_version


def create_node_groupings(
    spark_dataframe: DataFrame, source_col: str, target_col: str, num_groups: int
) -> DataFrame:
    """
    Create node groupings for parallel ingest into Neo4j.
    Add `source_group`, `target_group` and `group` columns to the Spark DataFrame identifying which groups the row belongs in.
    `group` is a concatenation of the source and target group values.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame to operate on.
    source_col : str
        The column indicating the relationship source id.
    target_col : str
        The column indicating the relationship target id.
    num_groups : int
        The desired number of groups to generate. The process may generate less groups as necessary.

    Returns
    -------
    DataFrame
        The Spark DataFrame with added columns `source_group`, `target_group` and `group`.
    """

    verify_spark_version(spark_session=spark_dataframe.sparkSession)

    # to create buckets
    # run over source and target INDEPENDENTLY
    # group by and count
    source_count_sdf = create_value_counts_dataframe(
        spark_dataframe=spark_dataframe, grouping_column=source_col
    )
    target_count_sdf = create_value_counts_dataframe(
        spark_dataframe=spark_dataframe, grouping_column=target_col
    )

    # iterate through the values in max -> min order ex: [{key: Amazon, value_count: 100000}, ...]
    # find most-empty bucket (num_groups) and place value in it and increment bucket value by value_count
    # # track with 2 separate hash maps
    source_groupings_sdf = create_value_groupings(
        value_counts_spark_dataframe=source_count_sdf,
        num_groups=num_groups,
        grouping_column=source_col,
    )
    target_groupings_sdf = create_value_groupings(
        value_counts_spark_dataframe=target_count_sdf,
        num_groups=num_groups,
        grouping_column=target_col,
    )

    final_sdf = spark_dataframe.join(
        other=source_groupings_sdf.withColumnRenamed("group", "source_group"),
        on=(spark_dataframe[source_col] == source_groupings_sdf.value),
        how="left",
    ).drop(source_groupings_sdf.value)
    final_sdf = final_sdf.join(
        other=target_groupings_sdf.withColumnRenamed("group", "target_group"),
        on=(spark_dataframe[target_col] == target_groupings_sdf.value),
        how="left",
    ).drop(target_groupings_sdf.value)

    final_sdf = final_sdf.drop("value")

    final_sdf = create_group_column_from_source_and_target_groups(final_sdf)

    return final_sdf



from pyspark.sql import DataFrame

from .batching import create_ingest_batches_from_groups
from .grouping import create_node_groupings


def group_and_batch_spark_dataframe(
    spark_dataframe: DataFrame, partition_col: str, num_groups: int
) -> DataFrame:
    """
    Create node groupings and batches for parallel ingest into Neo4j.
    Add `group` and `batch` columns to the Spark DataFrame identifying which grous and batch the row belongs in.
    `group` is a concatenation of the source and target group values.
    `group` and `batch` are utilized during ingestion.

    Parameters
    ----------
    spark_dataframe : DataFrame
        The Spark DataFrame to operate on.
    partition_col : str
        The desired column to partition on.
    num_groups : int
        The desired number of groups to generate. The process may generate less groups as necessary.

    Returns
    -------
    DataFrame
        The Spark DataFrame with added columns `group` and `batch`.
    """

    grouped_sdf = create_node_groupings(
        spark_dataframe=spark_dataframe,
        partition_col=partition_col,
        num_groups=num_groups,
    )
    return create_ingest_batches_from_groups(spark_dataframe=grouped_sdf)
